import unittest

from rxxxt.elements import El
from rxxxt.component import Component
from rxxxt.events import NavigateOutputEvent
from rxxxt.page import default_page
from rxxxt.session import Session, SessionConfig
from rxxxt.state import JWTStateResolver

session_config = SessionConfig(page_facotry=default_page, state_resolver=JWTStateResolver(b"deez"), persistent=False)

class TestSession(unittest.IsolatedAsyncioTestCase):
  async def test_state_cell_update(self):
    class Main(Component):
      def render(self):
        return El.div(content=[self.context.path])

    async with Session(session_config, Main()) as session:
      session.set_location("/hello-world")
      await session.init(None)
      update1 = await session.render_update(True, True)
      self.assertIn("/hello-world", update1.html_parts[0])

    async with Session(session_config, Main()) as session:
      await session.init(update1.state_token)
      session.set_location("/world-hello")
      await session.handle_events(()) # this should not matter but we want to match the App flow
      await session.update()
      update2 = await session.render_update(True, True)
      self.assertIn("/world-hello", update2.html_parts[0])

  async def test_event_deduplication(self):
    class Main(Component):
      def render(self):
        self.context.navigate("/hello-world")
        self.context.navigate("/world-hello")
        self.context.navigate("/hello-world")
        return El.div()

    async with Session(session_config, Main()) as session:
      session.set_location("/")
      await session.init(None)
      update = await session.render_update(True, True)
      self.assertEqual(update.events, (
        NavigateOutputEvent(location = "/hello-world"),
        NavigateOutputEvent(location = "/world-hello")
      ))

if __name__ == "__main__":
  _ = unittest.main()



from rxxxt.elements import Element, CustomAttribute, ElementContent, HTMLAttributeValue, HTMLAttributes, HTMLFragment, HTMLVoidElement, \
  HTMLElement, KeyedElement, WithRegistered, lazy_element, TextElement, UnescapedHTMLElement, El, VEl, ElementFactory
from rxxxt.component import EventHandler, event_handler, HandleNavigate, Component
from rxxxt.page import PageFactory, default_page, PageBuilder
from rxxxt.execution import State, Context
from rxxxt.app import App
from rxxxt.router import router_params, Router
from rxxxt.state import local_state, global_state, context_state, local_state_box, global_state_box, context_state_box
from rxxxt.helpers import class_map, match_path

__all__ = [
  "Element", "CustomAttribute", "ElementContent", "HTMLAttributeValue", "HTMLAttributes", "HTMLFragment", "HTMLVoidElement",
    "HTMLElement", "KeyedElement", "WithRegistered", "lazy_element", "TextElement", "UnescapedHTMLElement", "El", "VEl", "ElementFactory",

  "EventHandler", "event_handler", "HandleNavigate", "Component",

  "PageFactory", "default_page", "PageBuilder",

  "State", "Context",

  "App",

  "router_params", "Router",

  "local_state", "global_state", "context_state", "local_state_box", "global_state_box", "context_state_box",

  "class_map", "match_path"
]



from abc import ABC, abstractmethod
from datetime import timedelta
import inspect
import os
import secrets
from typing import Callable, Generic, get_origin, Any
from collections.abc import Awaitable
from pydantic import TypeAdapter, ValidationError

from rxxxt.cell import SerilializableStateCell
from rxxxt.component import Component
from rxxxt.execution import Context, State
from rxxxt.helpers import T, JWTError, JWTManager

StateDataAdapter = TypeAdapter(dict[str, str])

class StateBox(Generic[T]):
  def __init__(self, key: str, state: State, inner: SerilializableStateCell[T]) -> None:
    super().__init__()
    self._key = key
    self._state = state
    self._inner = inner

  def __enter__(self): return self
  def __exit__(self, *_): self.update()

  @property
  def key(self): return self._key

  @property
  def value(self): return self._inner.value

  @value.setter
  def value(self, v: T):
    self._inner.value = v
    self.update()

  def update(self): self._state.request_key_updates({ self._key })

class StateBoxDescriptorBase(Generic[T]):
  def __init__(self, state_key_producer: Callable[[Context, str], str], default_factory: Callable[[], T], state_name: str | None = None) -> None:
    self._state_name = state_name
    self._default_factory = default_factory
    self._state_key_producer = state_key_producer

    native_types = (bool, bytearray, bytes, complex, dict, float, frozenset, int, list, object, set, str, tuple)
    if default_factory in native_types or get_origin(default_factory) in native_types:
      self._val_type_adapter: TypeAdapter[Any] = TypeAdapter(default_factory)
    else:
      sig = inspect.signature(default_factory)
      self._val_type_adapter = TypeAdapter(sig.return_annotation)

  def __set_name__(self, owner: Any, name: str):
    if self._state_name is None:
      self._state_name = name

  def _get_box(self, context: Context) -> StateBox[T]:
    if not self._state_name: raise ValueError("State name not defined!")
    key = self._state_key_producer(context, self._state_name)

    if (cell:=context.state.get_key_cell(key)) is not None:
      if not isinstance(cell, SerilializableStateCell):
        raise ValueError(f"Cell is not serializable for key '{key}'!")
      return StateBox(key, context.state, cell)

    old_str_value = context.state.get_key_str(key)
    if old_str_value is None:
      value = self._default_factory()
    else:
      value = self._val_type_adapter.validate_json(old_str_value)
    cell = SerilializableStateCell(value, lambda v: self._val_type_adapter.dump_json(v).decode("utf-8"))
    context.state.set_key_cell(key, cell)
    return StateBox(key, context.state, cell)

class StateBoxDescriptor(StateBoxDescriptorBase[T]):
  def __get__(self, obj: Any, objtype: Any=None):
    if not isinstance(obj, Component):
      raise TypeError("StateDescriptor used on non-component!")

    box = self._get_box(obj.context)
    obj.context.subscribe(box.key)

    return box

class StateDescriptor(StateBoxDescriptorBase[T]):
  def __set__(self, obj: Any, value: Any):
    if not isinstance(obj, Component):
      raise TypeError("StateDescriptor used on non-component!")

    box = self._get_box(obj.context)
    box.value = value

  def __get__(self, obj: Any, objtype: Any=None):
    if not isinstance(obj, Component):
      raise TypeError("StateDescriptor used on non-component!")

    box = self._get_box(obj.context)
    obj.context.subscribe(box.key)

    return box.value

def get_global_state_key(_context: Context, name: str):
  return f"global;{name}"

def get_local_state_key(context: Context, name: str):
  return f"#local;{context.sid};{name}"

def get_context_state_key(context: Context, name: str):
  state_key = None
  exisiting_keys = context.state.keys
  for sid in context.stack_sids:
    state_key = f"#context;{sid};{name}"
    if state_key in exisiting_keys:
      return state_key
  if state_key is None: raise ValueError(f"State key not found for context '{name}'!")
  return state_key # this is just the key for context.sid

def local_state(default_factory: Callable[[], T], name: str | None = None):
  return StateDescriptor(get_local_state_key, default_factory, state_name=name)

def global_state(default_factory: Callable[[], T], name: str | None = None):
  return StateDescriptor(get_global_state_key, default_factory, state_name=name)

def context_state(default_factory: Callable[[], T], name: str | None = None):
  return StateDescriptor(get_context_state_key, default_factory, state_name=name)

def local_state_box(default_factory: Callable[[], T], name: str | None = None):
  return StateBoxDescriptor(get_local_state_key, default_factory, state_name=name)

def global_state_box(default_factory: Callable[[], T], name: str | None = None):
  return StateBoxDescriptor(get_global_state_key, default_factory, state_name=name)

def context_state_box(default_factory: Callable[[], T], name: str | None = None):
  return StateBoxDescriptor(get_context_state_key, default_factory, state_name=name)

class StateResolverError(BaseException): pass

class StateResolver(ABC):
  @abstractmethod
  def create_token(self, data: dict[str, str], old_token: str | None) -> str | Awaitable[str]: pass
  @abstractmethod
  def resolve(self, token: str) -> dict[str, str] | Awaitable[dict[str, str]]: pass

class JWTStateResolver(StateResolver):
  def __init__(self, secret: bytes, max_age: timedelta | None = None, algorithm: str = "HS512") -> None:
    super().__init__()
    self._jwt_manager = JWTManager(secret, timedelta(days=1) if max_age is None else max_age, algorithm)

  def create_token(self, data: dict[str, str], old_token: str | None) -> str:
    try: return self._jwt_manager.sign({ "d": data })
    except JWTError as e: raise StateResolverError(e)

  def resolve(self, token: str) -> dict[str, str]:
    try:
      payload = self._jwt_manager.verify(token)
      return StateDataAdapter.validate_python(payload["d"])
    except (ValidationError, JWTError) as e: raise StateResolverError(e)

def default_state_resolver() -> JWTStateResolver:
  """
  Creates a JWTStateResolver.
  Uses the environment variable `JWT_SECRET` as its secret, if set, otherwise creates a new random, temporary secret.
  """

  jwt_secret = os.getenv("JWT_SECRET", None)
  if jwt_secret is None: jwt_secret = secrets.token_bytes(64)
  else: jwt_secret = jwt_secret.encode("utf-8")
  return JWTStateResolver(jwt_secret)



import hmac
from inspect import isawaitable
from typing import Callable, ParamSpec, TypeVar, cast, Any, Literal
from collections.abc import Awaitable
import re, functools, hashlib, io, base64, json
from datetime import datetime, timedelta, timezone

from pydantic import BaseModel, TypeAdapter

T = TypeVar("T")
FNP = ParamSpec('FNP')
FNR = TypeVar('FNR')

async def to_awaitable(fn: Callable[..., T | Awaitable[T]], *args: Any, **kwargs: Any) -> T:
  result = fn(*args, **kwargs)
  if isawaitable(result): result = await result
  return cast(T, result)

def class_map(map: dict[str, bool]):
  return " ".join([ k for k, v in map.items() if v ])

_RE_PATH_PARAM_DEF = re.compile(r"\{([^\*\}]*)(\*)?\}")
_RE_PATH_VALID = re.compile(r"[A-Za-z0-9._~\-\/]*")
_RE_PATH_PART_VALID = re.compile(r"[A-Za-z0-9._~\-]*")

@functools.lru_cache()
def _compile_matcher(pattern: str, re_flags: int):
  re_parts: list[str] = []
  index = 0
  for m in _RE_PATH_PARAM_DEF.finditer(pattern):
    segment = pattern[index:m.start()]
    if not _RE_PATH_VALID.fullmatch(segment):
      raise ValueError(f"path segment '{segment}' in '{pattern}' is invalid!")
    re_parts.append(re.escape(segment))

    if str.isidentifier(m.group(1)):
      if m.group(2) == "*": re_parts.append(f"(?P<{m.group(1)}>{_RE_PATH_VALID.pattern})")
      else: re_parts.append(f"(?P<{m.group(1)}>{_RE_PATH_PART_VALID.pattern})")
    elif m.group(1) == "":
      if m.group(2) == "*": re_parts.append(f"({_RE_PATH_VALID.pattern})")
      else: re_parts.append(f"({_RE_PATH_PART_VALID.pattern})")
    else:
      raise ValueError(f"'{m.group(1)}' is not a valid part name in '{pattern}'!")

    index = m.end()

  final_segment = pattern[index:]
  if not _RE_PATH_VALID.fullmatch(final_segment):
    raise ValueError(f"path segment '{final_segment}' in '{pattern}' is invalid!")
  re_parts.append(re.escape(final_segment))

  pat_re = re.compile("".join(re_parts), re_flags)

  def _matcher(path: str) -> dict[str, str] | None:
    match = pat_re.fullmatch(path)
    if match is None: return None
    else: return match.groupdict()
  return _matcher

def match_path(pattern: str, path: str, re_flags: int = re.IGNORECASE):
  return _compile_matcher(pattern, re_flags)(path)

class JWTError(Exception): pass

def _jwt_b64url_encode(value: bytes | bytearray):
  return base64.urlsafe_b64encode(value).rstrip(b"=")
def _jwt_encode_json(obj: Any):
  return _jwt_b64url_encode(json.dumps(obj).encode())
def jwt_b64url_decode(value: bytes | bytearray):
  return base64.urlsafe_b64decode(value + b"=" * (4 - len(value) % 4))

class JWTManager:
  class JWTHeader(BaseModel):
    typ: Literal["JWT"]
    alg: str

  JWTPayloadAdapter = TypeAdapter(dict[str, Any])

  class JWTPayloadValidations(BaseModel):
    exp: int

    def is_valid(self):
      expires_dt = datetime.fromtimestamp(self.exp, timezone.utc)
      return expires_dt >= datetime.now(tz=timezone.utc)

  def __init__(self, secret: bytes, max_age: timedelta, algorithm: str = "HS512") -> None:
    super().__init__()
    self._secret = secret
    self._max_age: timedelta = max_age
    self._algorithm = algorithm
    self._digest = { "HS256": hashlib.sha256, "HS384": hashlib.sha384, "HS512": hashlib.sha512 }[algorithm]
    self._jwt_header = _jwt_encode_json({ "typ": "JWT", "alg": self._algorithm }) + b"."

  def sign(self, extra_fields: dict[str, Any]):
    try:
      expires_at = int((datetime.now(tz=timezone.utc) + self._max_age).timestamp())
      stream = io.BytesIO()
      _ = stream.write(self._jwt_header)
      _ = stream.write(_jwt_encode_json({ "exp": expires_at, **extra_fields }))
      signature = hmac.digest(self._secret, stream.getvalue(), self._digest)
      _ = stream.write(b".")
      _ = stream.write(_jwt_b64url_encode(signature))
      return stream.getvalue().decode()
    except Exception as e:
      if not isinstance(e, JWTError): raise JWTError(e)
      else: raise e

  def verify(self, token: str):
    try:
      parts = token.encode().split(b".")
      if len(parts) != 3: raise JWTError("invalid format (expected 3 parts)")

      header = JWTManager.JWTHeader.model_validate_json(jwt_b64url_decode(parts[0]))
      if header.alg != self._algorithm: raise JWTError("invalid algorithm in header")

      ref_signature = hmac.digest(self._secret, parts[0] + b"." + parts[1], self._digest)
      if not hmac.compare_digest(jwt_b64url_decode(parts[2]), ref_signature):
        raise JWTError("invalid JWT signature!")

      full_payload = JWTManager.JWTPayloadAdapter.validate_json(jwt_b64url_decode(parts[1]))
      if not JWTManager.JWTPayloadValidations.model_validate(full_payload).is_valid():
        raise JWTError("token expired")

      full_payload.pop("exp", None)
      return full_payload
    except Exception as e:
      if not isinstance(e, JWTError): raise JWTError(e)
      else: raise e



import typing
from rxxxt.elements import El, Element, ElementFactory, KeyedElement
from rxxxt.component import Component
from rxxxt.helpers import match_path
from rxxxt.state import context_state

def router_params(): return context_state(dict[str, str], "*rp*")

class Router(ElementFactory):
  class RoutedComponent(Component):
    params = router_params()

    def __init__(self, routes: tuple[tuple[str, ElementFactory], ...]):
      super().__init__()
      self._routes = routes
      self._selected_match: tuple[int, ElementFactory, dict[str, str]] | None = None

    async def on_before_update(self) -> None:
      self._selected_match = self._get_current_match()
      self.params = typing.cast(dict[str, str], dict()) if self._selected_match is None else self._selected_match[2]

    def render(self) -> Element:
      if self._selected_match is None:
        return El.h1(content=["Not found!"])
      else:
        return KeyedElement(str(self._selected_match[0]), self._selected_match[1]())

    def _get_current_match(self):
      path = self.context.path
      for idx, (pattern, element_factory) in enumerate(self._routes):
        if (m:=match_path(pattern, path)) is not None:
          return (idx, element_factory, m)
      return None

  def __init__(self) -> None:
    self._routes: list[tuple[str, ElementFactory]] = []

  def add_router(self, router: 'Router'): self._routes.extend(router._routes)
  def add_route(self, pattern: str, element_factory: ElementFactory): self._routes.append((pattern, element_factory))
  def route(self, pattern: str):
    def _inner(fn: ElementFactory):
      self.add_route(pattern, fn)
      return fn
    return _inner

  def __call__(self) -> Element: return Router.RoutedComponent(tuple(self._routes))



import unittest
from rxxxt.elements import El
from rxxxt.helpers import match_path
from rxxxt.router import Router
from tests.helpers import element_to_node, render_node

class TestMatcher(unittest.TestCase):
  def test_basic(self):
    pattern ="/hello/{one}/{all*}"
    self.assertDictEqual(match_path(pattern, "/hello/world/1337/42") or {}, { "one": "world", "all": "1337/42" })
    self.assertDictEqual(match_path(pattern, "/hello/world/1337") or {}, { "one": "world", "all": "1337" })
    self.assertIsNone(match_path(pattern, "/hello/1337"))

  def test_unnamed(self):
    self.assertEqual(match_path("/{*}", "/hello/world/1337"), {})
    self.assertEqual(match_path("/{*}/{one}", "/hello/world/1337"), { "one": "1337" })
    self.assertIsNone(match_path("/{}/{one}", "/hello/world/1337"))
    self.assertEqual(match_path("/{}/{}/{one}", "/hello/world/1337"), { "one": "1337" })
    self.assertEqual(match_path("/{}/{path*}", "/hello/world/1337"), { "path": "world/1337" })

class TestRouter(unittest.IsolatedAsyncioTestCase):
  async def test_basic(self):
    router = Router()
    router.add_route("/hello", lambda: El.div(content=["hello"]))
    router.add_route("/world", lambda: El.div(content=["world"]))

    comp = router()
    node = element_to_node(comp)
    node.context.state.update_state_strs({ "!location": "/hello" })
    await node.expand()
    self.assertEqual(render_node(node), "<div>hello</div>")

    node.context.state.update_state_strs({ "!location": "/world" })
    await node.update()
    self.assertEqual(render_node(node), "<div>world</div>")

    node.context.state.update_state_strs({ "!location": "/no" })
    await node.update()
    self.assertEqual(render_node(node), "<h1>Not found!</h1>")

    await node.destroy()

  async def test_var_path(self):
    router = Router()
    router.add_route("/var/{value}", lambda: El.div(content=["var1"]))
    router.add_route("/var/{a}/{b}", lambda: El.div(content=["var2"]))

    @router.route("/{path*}")
    def _(): return El.div(content=["not found"])

    comp = router()
    node = element_to_node(comp)
    node.context.state.update_state_strs({ "!location": "/hello" })
    await node.expand()
    self.assertEqual(render_node(node), "<div>not found</div>")

    node.context.state.update_state_strs({ "!location": "/var/1" })
    await node.update()
    self.assertEqual(render_node(node), "<div>var1</div>")

    node.context.state.update_state_strs({ "!location": "/var/1/2" })
    await node.update()
    self.assertEqual(render_node(node), "<div>var2</div>")

    await node.destroy()

if __name__ == "__main__":
  _ = unittest.main()
